/*******************************************************************************
 * BerthaFunctions.c
 * target: PushpinV3
 *
 * Josh Lifton
 * lifton@media.mit.edu
 * MIT Media Lab
 * Copyright 2002 all rights reserved.
 *
 * Last modified 07MAR2002 by lifton.
 *
 * These functions are for Bertha's internal use only.
 *******************************************************************************/

#include <c8051F000.h>
#include <PushpinHardwareV3.h>
#include <Bertha.h>
#include <BerthaFunctionPrototypes.h>
#include <BerthaGlobals.h>
#include <stdlib.h>
#include <string.h>


/*
 * Configure and enable analog to digital converter.  Select channel 0 as default.
 */
//void enableADC(void) {
//	REF0CN |= 0x03;		// Enable VREF and ADC/DAC bias.
//	AMX0CF = 0x00;		// Single-ended inputs (not differential).
//	AMX0SL = 0x00;		// Select Analog Input 0 (AIN0).
//	ADC0CF &= ~(0x07);	// Gain 1.
//	ADC0CF |= 0xE0;		// Conversion clock to system clock / 16.
//	ADLJST = 0;			// Right justification.
//	ADCTM = 1;			// Tracking is software controlled
//						// necessary for software control through ADBUSY 		
//	ADCEN = 1;			// Enable ADC.
//}

/*
 * Returns the current 10-bit value read from the specified ADC channel.  Assumes 
 * the data format of the ADC is right justified.  There are two issues to keep 
 * in mind when using the ADC.  First, there is some settling time required 
 * after changing channels.  If the impedance coming into the ADC channel is high, 
 * the RC time constant will not be negligible.  In particular, the voltage 
 * dividers used to identify the various communication and expansion modules are
 * typically high impedance.  Thus, some amount of delay may be required between 
 * switching ADC channels and detecting the communication or expansion modules.
 * The other issue is perhaps more critical.  It appears that the first ADC 
 * reading taken after changing channels comes from the previous channel.  Why 
 * this is so is not known, although one possibility is that the analog mux 
 * controlling access to the ADC doesn't actually change channels until a reading
 * is taken.  That is, changing the AMX0SL doesn't switch the mux in itself, 
 * but only lets the ADC know which channel to switch to when a reading is 
 * started (ADBUSY = 1).  Thus, that the first reading appears to be from the 
 * previous channel is just a result of the ADC being much faster than the RC 
 * time constant associated with the input.
 * Update: see Cygnal application note.
 */
unsigned int getADCValue(unsigned char channel) {
	AddressableInt ADCvalue;
	if (enableADCChannel(channel)) {
		ADBUSY = 1;				// Start ADC tracking and conversion.
		while (!ADCINT);		// Wait for conversion to end.
		ADCINT = 0;				// Clear interrupt.
		ADCvalue.Char[0] = ADC0H;
		ADCvalue.Char[1] = ADC0L;
		return ADCvalue.Int;
	}
	return 0;
}

/*
 * Returns '1' if the requested channel has been switched to and the ADC 
 * is ready to sample.  Returns '0' if the channel requested doesn't exist.
 */
unsigned char enableADCChannel(unsigned char channel) {
	if (channel < NUMBEROFADCCHANNELS) {
		REF0CN |= 0x03;		// Enable VREF and ADC/DAC bias.
		AMX0CF = 0x00;		// Single-ended inputs (not differential).
		ADC0CF &= ~(0x07);	// Gain 1.
		ADC0CF |= 0xE0;		// Conversion clock to system clock / 16.
		ADLJST = 0;			// Right justification.
		ADCTM = 1;			// Tracking is software controlled
							// necessary for software control through ADBUSY 		
		ADCEN = 1;			// Enable ADC.
		if (AMX0SL != channel) {
			AMX0SL = channel;	// Set analog mux channel select.
			ADBUSY = 1;			//| Take one reading so as to force the mux to 
			ADCINT = 0;			//|
			while (!ADCINT);	//| switch and let input settle according to 
			ADCINT = 0;			//| its RC time constant.
		}
		return 1;
	}
	return 0;
}

void monitorADCWindow(unsigned char channel, unsigned int lowerBound, unsigned int upperBound) {
	if (enableADCChannel(channel)) {
		ADCWindowGreaterThan = lowerBound;	// Set the lower bound of the window.
		ADCWindowLessThan = upperBound;		// Set the upper bound of the window.
		ADC0CN |= 0x04;		// Conversions (and tracking) started on timer 3 overflows.
		clearADCWindowInterruptFlag();
		enableADCWindowInterrupt();
		enableTimer3();
	}
}

void ADCWindowInterrupt(void) interrupt 8 {
	/*
	 * Add appropriate handler call here.
	 * E.g. LDRAboveThreshold();
	 */
	clearADCWindowInterruptFlag();
}

void configurePins(void) {
	PRT0CF = 0xFF;	// Set all of port 0 to push-pull outputs.
	PRT1CF = 0xFF;	// Set all of port 1 to push-pull outputs.
	PRT2CF = 0xFF;	// Set all of port 2 (not pinned out) to push-pull outputs.
	XBR0 |= 0x04;	// UART TX and RX on P0.0 and P0.1.	
	XBR2 |= 0x40;	// Enable crossbar. 
	XBR2 |= 0x80;	// Disable weak pull-ups.
}

void disableWatchdogTimer(void) {
	WDTCN = 0xde;	// This register must be set first.
	WDTCN = 0xad;	// This register must be set second.
}

void enableExternalClock(void) {
	CKCON |= 0x10;				// Don't divide by 12 for base clock; ie, T1M=1.
	OSCXCN = 0x67;				// Start external oscillator > 6.74MHz (using a 22.118MHz crystal)
	delay(0xffff);				// Wait more than a millisecond.
	while (!(OSCXCN & 0x80)) {} // Wait until external oscillator is stable.
	OSCICN |= 0x80;				// Check for missing clock.
	OSCICN |= 0x08;				// Use external crystal.
	FLSCL = ((FLSCL & 0xF0) | 0x09);	// Set flash scalar for 22Mhz clock.
}

void initializeHardware(void) {
	disableWatchdogTimer();
	enableExternalClock();
	configurePins();
	statusLED = 0;
}

/*
 * A wrapper for the srand() function provided in stdlib.h.
 */
void setRandomSeed(int seed) {
	srand(seed);
}

/* 
 * Copies a contiguous block of xdata memory to flash memory.  
 * Disables interrupts during sensitive tasks and then restores interrupt state.
 */
void copyXdataToFlash(unsigned char code * to, unsigned char xdata * from, unsigned int size) {
	unsigned int i;
	unsigned char xdata c;
	unsigned char interruptState = EA;
	if ((to + size < FlashSize) && (from + size < XdataSize)) {
		disableGlobalInterrupts();
		for (i=0; i<size; i++) {
			c = *(from + i);
			enableFlashWrite();
			*((unsigned char xdata *) to) = c;
			disableFlashWrite();
			to++;
		}
		EA = interruptState;
	}			
}

void saveOSState(OSState xdata * state) {
	clearFlashMemory((unsigned char code *) &osState, sizeof(OSState));
	copyXdataToFlash((unsigned char code *) &osState, (unsigned char xdata *) state, sizeof(OSState));
}

/*
 * A blocking function that flashes the LED located on the processing
 * module a certain number of times at a certain interval, both given 
 * as arguments. To be replaced by use of a timer.
 */
void flashLED(unsigned int times, unsigned int interval) {
	while(times--) {
		statusLED = 1;
		delay(interval);
		statusLED = 0;
		delay(interval);
	}
}

/* 
 * A blocking function that simply waits an amount of time roughly 
 * proportional to the value of the argument.
 */
void delay(unsigned int interval) {
	while(interval--);
}