#include <c8051F000.h>
#include <PushpinHardwareV3.h>
#include <Bertha.h>
#include <BerthaFunctionPrototypes.h>

/*
   meme : the problem described below is solved by explicitely casting all #defined constants when 
   they are defined.  Otherwise, the compiler treats them as signed and must call special libraries 
   to deal with them.

   meme : be sure to include all necessary covert functions included by the compiler in the correct
   memory space.  In particular, only the ?C?SLCMP module seems to be needed.  This module is 
   included in the ?C?LIB_CODE segment along with all other library functions called upon by the
   code.  Since ?C?LIB_CODE will change according to the needs of all the code, not just the code 
   in the OSOverwrite segment, ?C?LIB_CODE shouldn't be included as part of the permanent update 
   code.  It is unclear whether it is possible to locate only ?C?SLCMP with the update code.  Search
   http://www.keil.com/discuss/ for Slarti Bartfast to find out if this question has been answered.
   For now, all the functions in ?C?LIB_CODE are placed with this segment.
*/

/*
   meme : the variable interruptState in the functions below should be optimized to be a register variable.
*/

void overwriteOS(unsigned char code * to, unsigned char code * from, unsigned int size) {
	disableGlobalInterrupts();
	clearFlashMemory(to, size); 
	copyFlashBlock(to, from, size);
	reboot();
}

/* 
 * Copies a contiguous block of flash memory to another location in flash memory.  
 * Disables interrupts during sensitive tasks and then restores interrupt state.
 */
void copyFlashBlock(unsigned char code * to, unsigned char code * from, unsigned int size) {
	unsigned char xdata * i;
	unsigned char xdata c;
	unsigned char interruptState = EA;
	if ((to + size < FlashSize) && (from + size < FlashSize)) {
		if (to < from) {
			for (i=(unsigned char xdata *) to; i<(unsigned char xdata *) (to+size); i++) {
				c = *((unsigned char code *) (((unsigned int) from) + i - ((unsigned int) to)));
				disableGlobalInterrupts();
				enableFlashWrite();
				*i = c;
				disableFlashWrite();
				EA = interruptState;
			}
		}
		else if (to > from) {
			for (i=(unsigned char xdata *) (to+size-1); i>(unsigned char xdata *) to; i--) {
				c = *((unsigned char code *) (((unsigned int) from) + i - ((unsigned int) to)));
				disableGlobalInterrupts();
				enableFlashWrite();
				*i = c;
				disableFlashWrite();
				EA = interruptState;
			}
		}
	}
}

/*
 * If the arguments to this function specify a valid block of flash 
 * memory, the smallest area of erasable flash containing that 
 * block will be cleared.  Since the smallest unit of erasable 
 * flash is 512 bytes, up to 1022 extra bytes of memory might be 
 * erased (511 bytes before the block and 511 bytes after the block.
 */
void clearFlashMemory(unsigned char code * startPtr, unsigned int size) {
	unsigned char interruptState = EA;
	/*
	 * Check that the size is not too large.
	 */
	if (size == 0 || size > FlashSize || ((unsigned int) startPtr) + size < ((unsigned int) startPtr)) {
		return;
	}
	/*
	 * Erase all but the last flash page within range.
	 */
	while (size > FlashPageSize) {
		disableGlobalInterrupts();
		enableFlashErase();
		*((unsigned char xdata *) startPtr) = 0xFF;
		disableFlashErase();
		EA = interruptState;
		size -= FlashPageSize;
		startPtr += FlashPageSize;
	}
	/*
	 * Erase the last flash page within range.
	 */
	disableGlobalInterrupts();
	enableFlashErase();
	*((unsigned char xdata *) startPtr) = 0xFF;
	disableFlashErase();
	EA = interruptState;
}